#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <syslog.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include "dhcp.h"
#include "dhcpparse.h"

int dhcp_find_option(u_int8_t *options, int len, int type)
{
	int offs = 0;
	for(offs = 0; offs < len; )
	{
		if(!options[offs])
			offs++;
		else if(options[offs] == 0xff)
			break;
		else if(options[offs] == type)
			return(offs);
		else
			offs += options[offs + 1] + 2;
	}
	return(-1);
}

char* dhcp_otostr(u_int8_t *data, int len, int suggestion)
{
	int b = 0;
	static char buffer[MAXBSIZE + 1];
	int blen = MAXBSIZE;
	
	
	switch(suggestion)
	{
		case DHCP_TYP_INTEGER:
			if(len != sizeof(int))
				suggestion = DHCP_TYP_ANY;
			break;
		case DHCP_TYP_IPLIST:
			if(len & 0x3)
				suggestion = DHCP_TYP_ANY;
			break;
		case DHCP_TYP_STRING:
		{
			int i;
			for(i = 0; i<len; i++)
			{
				char c = data[i];
				if((!isascii(c)) || c == 0)
				{
					suggestion = DHCP_TYP_ANY;
					break;
				}
			}
			break;
		}
		case DHCP_TYP_NULSTRING:
		{
			int i;
			for(i = 0; i<len; i++)
			{
				char c = data[i];
				if(c == 0) break;
				if(!isascii(c))
				{
					suggestion = DHCP_TYP_ANY;
					break;
				}
			}
			break;
		}
	}
	
	
	switch(suggestion)
	{
		case DHCP_TYP_STRING:
		case DHCP_TYP_NULSTRING:
			if(blen < (len + 2))
				return(NULL);
			buffer[b++] = 's';
			memcpy(buffer + b, data, len);
			b += len;
			break;
		case DHCP_TYP_INTEGER:
			if(blen < 3)
				return(NULL);
			buffer[b++] = 'i';
			b += snprintf(buffer, blen - b, "%d", ntohl(*((int*) (data))));
			break;
		case DHCP_TYP_IPLIST:
		{
			int ac = len >> 2, i;
			if(blen < (17 * ac + 2))
				return(NULL);
			buffer[b++] = 'a';
			for(i = 0; i<ac; i++)
				b += snprintf(buffer + b, blen - b, "%s%d.%d.%d.%d", i ? ", " : "", data[i*4], data[1 + i*4], data[2 + i*4], data[3 + i*4]);
			break;
		}
		default:
		{
			int i;
			if(blen < (3 * len + 2))
				return(NULL);
			
			buffer[b++] = 'h';
			for(i = 0; i<len; i++)
				b += snprintf(buffer + b, blen - b, "%s%02X", i ? ":" : "", data[i]);
			break;
		}
	}
	buffer[b] = 0;
	return(buffer);
}

int dhcp_strtoo(u_int8_t *data, char *string, int len, int suggestion)
{
	int l;
	if(!*string)
		return(0);

	switch(*string)
	{
		case 'a': suggestion = DHCP_TYP_IPLIST;	string++; break;
		case 'i': suggestion = DHCP_TYP_INTEGER;	string++; break;
		case 'h': suggestion = DHCP_TYP_HEXSTRING;	string++; break;
		case 's': suggestion = DHCP_TYP_STRING;	string++; break;
	}
	l = strlen(string);

	if(suggestion == DHCP_TYP_ANY)
	{
		int i;
		char *str;
		
		/* integer */
		strtol(string, &str, 0);
		if(!*str)
		{
			suggestion = DHCP_TYP_INTEGER;
			goto proceed;
		}
		
		/* hex string: 0xbadcafe0 or badcafe0 */
		if(!(l & 1))
		{
			for(i = strncmp(string, "0x", 2) ? 0 : 2; string[i]; i++)
			{
				if(!isxdigit(string[i]))
					break;
			}
			if(!string[i])
			{
				suggestion = DHCP_TYP_HEXSTRING;
				goto proceed;
			}
		}
		
		/* ip list: 1.2.3.4, 2.3.4.5, .... */
		
		for(str = string, i = 0; ; )
		{
			struct in_addr a;
			int res;
			char c;
			char *comma;
			
			while(*str == '\t' || *str == ' ') str++;
			if(!*str) break;
			
			comma = index(str, ',');
			if(comma)
			{
				c = *comma;
				*comma = 0;
			}
	
			res = inet_aton(str, &a);
			
			if(comma)
				*comma = c;
			
			if(!res)
			{
				i = 0;
				break;
			}
			else
				i++;
			
			if(comma)
				str = comma + 1;
			else
				break;
		}
		
		if(i > 0)
		{
			suggestion = DHCP_TYP_IPLIST;
			goto proceed;
		}

		suggestion = DHCP_TYP_STRING;	
	}		
proceed:

	switch(suggestion)
	{
		default:
		case DHCP_TYP_STRING:
			if(len < l) return(-1);
			memcpy(data, string, l);
			return(l);
		case DHCP_TYP_INTEGER:
		{
			char *eptr;
			
			if(len < sizeof(u_int32_t)) return(-1);
			*((u_int32_t*) data) = htonl(strtol(string, &eptr, 0));
			if(*eptr)
				return(-1);
			return(sizeof(u_int32_t));
		}
		case DHCP_TYP_HEXSTRING:
		{
			int i, z = 0;
			int hi = -1, lo;
			for(i = 0; i<l; i++)
			{
				int c;

				c = string[i];

				if(!isxdigit(c))
					continue;
				
				if(c >= 'a')
					c -= 'a' - 10;
				else if(c >= 'A')
					c -= 'A' - 10;
				else
					c -= '0';
					
				if(hi == -1)
					hi = c;
				else
				{
					lo = c;
					if(len < 1) return(-1);
					hi = (hi << 4) | lo;
					data[z++] = hi;
					len--;
					hi = -1;
				}
			}
			if(hi != -1) 
				return(-1);
			else
				return(z);
		}
		case DHCP_TYP_IPLIST:
		{
			char *str;
			int i, res;
			for(str = string, i = 0; ; )
			{
				struct in_addr a;
				char c;
				char *comma;
			
				while(*str == '\t' || *str == ' ') str++;
				if(!*str) break;
			
				comma = index(str, ',');
				if(comma)
				{
					c = *comma;
					*comma = 0;
				}
		
				res = inet_aton(str, &a);
			
				if(comma)
					*comma = c;
			
				if(!res)
					return(-1);
				else
				{
					if(len < 4) return(-1);
					*((u_int32_t*) (data + i)) = a.s_addr;

					i += 4;
					len -= 4;
				}
			
				if(comma)
					str = comma + 1;
				else
					break;
			}
			return(i);
		}
	}
}

static struct dhcpoption common_options[] = {
				{DHCP_OPTION_TYPE,	DHCP_TYP_INTEGER,	"message type",	"Message type"},
				{DHCP_OPTION_SUBNETMASK,DHCP_TYP_IPLIST,	"subnet mask",	"Subnet mask"},
				{DHCP_OPTION_TIMEOFFSET,DHCP_TYP_INTEGER,	"time offset",	"Time offset"},
				{DHCP_OPTION_ROUTERS,	DHCP_TYP_IPLIST,	"router",	"Routers"},
				{DHCP_OPTION_TIMESERVERS,DHCP_TYP_IPLIST,	"time server",	"Time servers"},
				{DHCP_OPTION_DNSSERVERS, DHCP_TYP_IPLIST,	"dns server",	"DNS servers"},
				{DHCP_OPTION_LOGSERVERS, DHCP_TYP_IPLIST,	"log server",	"Log servers"},
				{DHCP_OPTION_HOSTNAME,	DHCP_TYP_STRING,	"hostname",	"Hostname"},
				{DHCP_OPTION_DOMAINNAME,DHCP_TYP_STRING,	"domainname",	"Domainname"},
				{DHCP_OPTION_ARPTIMEOUT,DHCP_TYP_INTEGER,	"arp timeout",	"ARP timeout"},
				{DHCP_OPTION_LEASETIME,	DHCP_TYP_INTEGER,	"lease time",	"Lease time"},
				{DHCP_OPTION_SERVERID,	DHCP_TYP_HEXSTRING,	"server id",	"Server ID"},
				{DHCP_OPTION_REQUESTEDADDR, DHCP_TYP_IPLIST,	"requested address",	"Requested address"},
				{DHCP_OPTION_TFTPSERVERNAME, DHCP_TYP_STRING,	"tftp server",	"TFTP servers"}};
				
                                      

struct dhcpoption * dhcp_getoptionbyid(int option)
{
	int i;
	for(i = 0; i < (sizeof(common_options) / sizeof(struct dhcpoption)); i++)
		if(common_options[i].option == option) return(common_options + i);
	return(NULL);
}

struct dhcpoption * dhcp_getoptionbyname(char* option)
{
	int i;
	for(i = 0; i < (sizeof(common_options) / sizeof(struct dhcpoption)); i++)
		if(!strcmp(common_options[i].name, option)) return(common_options + i);
	return(NULL);
}

int dhcp_describe(struct dhcp_requesthdr *r, struct sockaddr_in peer, int length, char *string, int len, char *mac)
{
	int pos = 0;
	u_int8_t *options;
	int optlen, offs;
	
	pos += snprintf(string + pos, len - pos, "source ip: a%s\n", inet_ntoa(peer.sin_addr));
	pos += snprintf(string + pos, len - pos, "source port: i%d\n", ntohs(peer.sin_port));
	pos += snprintf(string + pos, len - pos, "flags: %s\n", dhcp_otostr((u_int8_t*) r->flags, 2, DHCP_TYP_HEXSTRING));
	pos += snprintf(string + pos, len - pos, "client hardware: %s\n", dhcp_otostr((u_int8_t*)r->chaddr, 6, DHCP_TYP_HEXSTRING));
	pos += snprintf(string + pos, len - pos, "transaction id: %s\n", dhcp_otostr((u_int8_t*)&r->xid, 4, DHCP_TYP_HEXSTRING));
	pos += snprintf(string + pos, len - pos, "requested address: %s\n", dhcp_otostr((u_int8_t*)r->ciaddr, 4, DHCP_TYP_IPLIST));
	pos += snprintf(string + pos, len - pos, "your address: %s\n", dhcp_otostr((u_int8_t*)r->yiaddr, 4, DHCP_TYP_IPLIST));
	pos += snprintf(string + pos, len - pos, "next server: %s\n", dhcp_otostr((u_int8_t*)r->siaddr, 4, DHCP_TYP_IPLIST));
	pos += snprintf(string + pos, len - pos, "relay: %s\n", dhcp_otostr((u_int8_t*)r->giaddr, 4, DHCP_TYP_IPLIST));
	if(r->file[0])
		pos += snprintf(string + pos, len - pos, "boot filename: %s\n", dhcp_otostr((u_int8_t*)r->file, 128, DHCP_TYP_NULSTRING));
	
	options = ((char*) r) + sizeof(struct dhcp_requesthdr);
	optlen = length - sizeof(struct dhcp_requesthdr);
	
//	do_log(LOG_INFO, "dhcp_describe(): Describing the message...");
	
	for(offs = 0; offs < optlen; )
	{
		if(!options[offs])
			offs++;
		else if(options[offs] == 0xff)
			break;
		else
		{
			struct dhcpoption * opt = dhcp_getoptionbyid(options[offs]);

//			do_log(LOG_INFO, "dhcp_describe(): Retrieving description for %d (%d, %d)...", options[offs], offs, optlen);

			if(opt)
				pos += snprintf(string + pos, len - pos, "%s: ", opt->name);
			else
				pos += snprintf(string + pos, len - pos, "%d: ", options[offs]);
			
			pos += snprintf(string + pos, len - pos, "%s\n", dhcp_otostr(options + offs + 2, options[offs + 1], opt ? opt->suggestion : DHCP_TYP_ANY));
			offs += options[offs + 1] + 2;
	
//			do_log(LOG_INFO, "dhcp_describe(): OK retrieving description for %d...", options[offs]);
		}
	}
	pos += snprintf(string + pos, len - pos, "end of message\n");
	string[pos] = 0;
	return(pos);
}

int dhcp_build(char *string, struct dhcp_requesthdr * msg, struct dhcp_requesthdr * r, int maxlen, struct in_addr *ina)
{
	char *c;
	u_int8_t *options;
	int optmax, pos = 0;
	char *colon;
		
	ina->s_addr = INADDR_ANY;
	
	assert(c = strdup(string));	/* will leak */
	
	memcpy(msg, r, sizeof(struct dhcp_requesthdr));

	msg->op = OP_BOOTREPLY;
	msg->hops = 0;
	memset(msg->secs, 0, sizeof(msg->secs));
	memset(msg->yiaddr, 0, sizeof(u_int32_t));
	memset(msg->siaddr, 0, sizeof(msg->siaddr));


	if(*((u_int32_t*) msg->giaddr))
		memset(msg->flags, 0, sizeof(msg->flags));
	memset(msg->giaddr, 0, sizeof(msg->giaddr));
	memset(msg->sname, 0, sizeof(msg->sname));
				
	memset(msg->file, 0, sizeof(msg->file));

	options = ((char*) msg) + sizeof(struct dhcp_requesthdr);
	optmax = maxlen - sizeof(struct dhcp_requesthdr);

	if(!*c)
	{
		do_log(LOG_WARNING, "dhcp_build(): The string is empty, refusing to bounce the message back!");
		return(-1);
	}

	while(*c)
	{
		char *nx;
		struct dhcpoption *opt;
		
		if((nx = index(c, '\n')))
			*(nx++) = 0;
		
		if(*c == '#') goto proceed;
		
		colon = index(c, ':');
		if(!colon)
		{
			do_log(LOG_WARNING, "dhcp_build(): Incorrect statement: %s, ignoring", c);
			goto proceed;
		}

		*colon = 0; colon++;
		while(*c == ' ' || *c == '\t') c++;
		while(*colon == ' ' || *colon == '\t') colon++;
		
//		do_log(LOG_INFO, "dhcp_build(): %s, %s", c, colon);
		
		if(!strcmp(c, "source address"))
		{
			if(dhcp_strtoo((u_int8_t*) ina->s_addr, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;			
		}
		else if(!strcmp(c, "next server"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->siaddr, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "flags"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->flags, colon, 2, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "client hardware"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->chaddr, colon, 12, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "transaction id"))
		{
			if(dhcp_strtoo((u_int8_t*) &msg->xid, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "requested address"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->ciaddr, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "your address"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->yiaddr, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "relay"))
		{
			if(dhcp_strtoo((u_int8_t*) msg->giaddr, colon, 4, DHCP_TYP_ANY) < 0)
				goto bad;
		}
		else if(!strcmp(c, "boot filename"))
		{
			int l;
			if((l = dhcp_strtoo((u_int8_t*) msg->file, colon, 128, DHCP_TYP_ANY)) < 0)
				goto bad;
			else
				msg->file[l] = 0;
			
		}
		else if((opt = dhcp_getoptionbyname(c)))
		{
			int l;
			l = dhcp_strtoo(options + pos + 2, colon, optmax - 2, DHCP_TYP_ANY);
			if(l < 0)
				goto bad;
//			fprintf(stderr, "saving %d at %d (size = %d)\n", opt->option, pos, l);
			options[pos] = opt->option;
			options[pos + 1] = l;
			pos += l + 2;
			optmax -= l + 2;
		}
		else
		{
			char *eptr;
			int opt, l;
			opt = strtol(c, &eptr, 0);
			if(*eptr)
			{
				do_log(LOG_WARNING, "dhcp_build(): %s: %s, unknown option, ignoring.", c, colon);
				goto proceed;
			}

			l = dhcp_strtoo(options + pos + 2, colon, optmax - 2, DHCP_TYP_ANY);
			if(l < 0)
				goto bad;

//			fprintf(stderr, "saving %d at %d (size = %d)\n", opt, pos, l);

			options[pos] = opt;
			options[pos + 1] = l;
			pos += l + 2;
			optmax -= l + 2;
		}
		
proceed:		
		if(!nx)
			break;
		c = nx;
	}
	
	if(optmax < 2) goto bad;
	options[pos++] = 0xff;
	return(pos + sizeof(struct dhcp_requesthdr));
	
bad:
	do_log(LOG_WARNING, "dhcp_build(): %s: %s, does not fit into a proper field or unable to convert into binary, aborting.", c, colon);
	return(-1);
}
