/*
 * LMSd
 *
 *  (C) Copyright 2002-2003 Mateusz 'mteg' Golicz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 *
 */

/* ten plik zawiera rozne przydatne funkcje, miedzy innymi sluzace do przetwarzania
   ciagow tekstowych, ale nie tylko */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "util.h"


void *amalloc(int size)
/* przydziel pamiec i upewnij sie, ze sie udalo */
{
	void * m;
	m = malloc(size);
	assert(m); /* jesli NULL to bedzie wybuch! */
	return(m);
}

void *arealloc(void *m, int size)
/* zrob realloc() i upewnij sie, ze sie udalo */
{
	m = realloc(m, size);
	assert(m); /* m == NULL -> BOOM ! */
	return(m);
}

unsigned char * str_replace(unsigned char * pattern, unsigned char * replacement, unsigned char * string, int stop)
/* zamien w stringu string stop wystapien pattern na replacement.
   jesli stop == 0 zamieniane sa wszystkie wystapienia
   string jest zwalniany za pomoca free() */
{
	unsigned char * occurence;
	int n = 0;
	
	while((occurence = strstr(string, pattern))) /* szukaj kolejnego wystapienia wzorca */
	{
		int l;
		unsigned char * new;
		
		*occurence = 0; /* w miejscu gdzie wystapil wzorzec - koniec ciagu */
		l = strlen(string); /* tyle bajtow bedzie trzeba skopiowac na pewno */
		
		occurence += strlen(pattern); /* etc, etc, ta funkcja nie jest jakos kosmicznie skomplikowana */
		l += strlen(occurence);
		l += strlen(replacement);
		
		new = (unsigned char*) malloc(l + 2);
		if(!new) return(NULL);
		
		snprintf(new, l + 1, "%s%s%s", string, replacement, occurence);
		free(string);
		string = new;
	
		n++;
		if(n == stop) break;
	}
	
	return(string);
	
}


unsigned char * str_concat(unsigned char * s1, unsigned char * s2)
/* funkcja laczy dwa ciagi s1 i s2. obydwa sa nastepnie zwalniane uzywajac free() */
{
	int l = strlen(s1) + strlen(s2) + 1;
	unsigned char * ret = amalloc(l);
	
	snprintf(ret, l, "%s%s", s1, s2);
	free(s1);
	free(s2);
	return(ret);
}

int get_ip_and_netmask(char * literal, u_int32_t * ip, u_int32_t * netmask, char *error, int errlen)
{
	char * slash;
	struct in_addr iip;
	struct in_addr imask;
	char * first = literal;
	
	assert(literal = strdup(literal));
	slash = index(literal, '/');
	
	if(slash)
	{
		char * eptr;
		int len;

		*slash = 0;
		slash ++;
		len = strtol(slash, &eptr, 10);

		if(*eptr)
		{
			if(!inet_aton(slash, &imask))
			{
				snprintf(error, errlen, "Invalid network mask in '%s'", first);
				return(0);
			}
		}
		else
		{
			unsigned long nmask = 0;
			while((len--) > 0)
			{
				nmask >>= 1;
				nmask |= 0x80000000;
			}
			imask.s_addr = htonl(nmask);
		}
	}
	else
	{
		imask.s_addr = 0xffffffff;				
	}
		
	if(!inet_aton(literal, &iip))
	{
		snprintf(error, errlen, "Invalid IP address in '%s'", first);
		return(0);
	}
	if(iip.s_addr & (~imask.s_addr))
	{
		snprintf(error, errlen, "Host bits set in '%s'", first);
		return(0);
	}

	free(literal);
	*ip = iip.s_addr;
	*netmask = imask.s_addr;
	return(1);
}

static inline int a2h(int x)
{
	if(x >= '0' && x <= '9') return(x - '0');
	if(x >= 'a' && x <= 'f') return(x - 'a' + 10);
	if(x >= 'A' && x <= 'F') return(x - 'A' + 10);
	return(-1);
}

int parse_mac(char * literal, char * hw)
{
	int i, n = 0;
	
	if(strlen(literal) == 12)
	{
		for(i = 0; i < 6; i++)
		{
			int q1 = a2h(literal[n++]);
			int q2 = a2h(literal[n++]);
			
			if(q1 == -1 || q2 == -1) return(0);
			hw[i] = (q1 << 4) | q2;
		}
		return(1);
	}
	else if(strlen(literal) == 14)
	{
		for(i = 0; i < 6; i++)
		{
			int q1;
			int q2;
			
			if(i == 2 || i == 4)
				if(literal[n++] != '.') return(0);

			q1 = a2h(literal[n++]);
			q2 = a2h(literal[n++]);
			
			if(q1 == -1 || q2 == -1) return(0);
			hw[i] = (q1 << 4) | q2;
		}
		return(1);
	}
	else if(strlen(literal) == 17)
	{
		for(i = 0; i < 6; i++)
		{
			int q1;
			int q2;
			
			if(i != 0)
				if(literal[n++] != ':') return(0);

			q1 = a2h(literal[n++]);
			q2 = a2h(literal[n++]);
			
			if(q1 == -1 || q2 == -1) return(0);
			hw[i] = (q1 << 4) | q2;
		}
		return(1);
	}
	return(0);
}

int parselong(char *str, char *error, int imin, int imax)
{
	char *eptr;
	int i;
	
	i = strtol(str, &eptr, 0);
	
	if(*eptr || (i < imin) || ((i > imax) && (imax > imin)))
	{
		fprintf(stderr, error, str);
		_exit(1);
	}
	return(i);
}
